package terminal;

import java.io.*;
import java.util.concurrent.*;

import com.googlecode.lanterna.TerminalPosition;
import com.googlecode.lanterna.input.*;
import com.googlecode.lanterna.terminal.*;

public class LispTerminal {

    // private static final String ANSI_RESET = "001B[0m";
    // private static final String ANSI_RED = "\u001B[31m";
    // private static final String ANSI_GREEN = "\u001B[32m";
    // private static final String ANSI_YELLOW = "\u001B[33m";
    // private static final String ANSI_PURPLE = "\u001B[35m";

    private IOSafeTerminal terminal;
    private boolean isFinished;
    private String currentLine;
    private TerminalPosition origin;
    private PipedOutputStream inputWriter;
    private PipedInputStream outputReader;
    private ExecutorService executor;

    LispTerminal(IOSafeTerminal terminal, PipedOutputStream inputWriter, PipedInputStream outputReader) {
        this.terminal = terminal;
        initialize(inputWriter, outputReader);
    }

    public LispTerminal(PipedOutputStream inputWriter, PipedInputStream outputReader) {
        try {
            Terminal unsafe = new DefaultTerminalFactory().createTerminal();
            this.terminal = IOSafeTerminalAdapter.createRuntimeExceptionConvertingAdapter(unsafe);
        } catch (IOException e) {
            // TODO
            e.printStackTrace();
        }

        initialize(inputWriter, outputReader);
    }

    private void initialize(PipedOutputStream inputWriter, PipedInputStream outputReader) {
        this.isFinished = false;
        this.currentLine = "";
        this.origin = terminal.getCursorPosition();
        this.inputWriter = inputWriter;
        this.outputReader = outputReader;
        this.executor = Executors.newFixedThreadPool(2);
    }

    public void run() {
        executor.execute(this::readInput);
        executor.execute(this::writeOutput);
        executor.shutdown();
    }

    public void readInput() {
        while (!isFinished) {
            handleKeyStroke(getKeyStroke());
            terminal.flush();

            try {
                Thread.sleep(1);
            } catch (InterruptedException e) {
                isFinished = true;
            }
        }
    }

    public void writeOutput() {
        try {
            for (int c = outputReader.read(); c != -1; c = outputReader.read()) {
                synchronized (this) {
                    terminal.setCursorVisible(false);

                    if (c == 'x') {
                        terminal.flush();
                        origin = terminal.getCursorPosition();
                        terminal.setCursorVisible(true);
                    } else {
                        terminal.putCharacter((char) c);
                    }
                }
            }
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        terminal.setCursorVisible(true);
        terminal.close();
    }

    private KeyStroke getKeyStroke() {
        KeyStroke keyStroke = null;

        try {
            keyStroke = terminal.pollInput();
        } catch (IllegalStateException e) {
            terminal.putCharacter('\n');
            System.exit(0);
        }

        return keyStroke;
    }

    private synchronized void handleKeyStroke(KeyStroke keyStroke) {
        if (keyStroke == null)
            return;

        KeyType keyType = keyStroke.getKeyType();

        if (keyStroke.isCtrlDown())
            doControlKeyCharacter(keyStroke);
        else if (keyType == KeyType.ArrowLeft)
            moveCursorLeft();
        else if (keyType == KeyType.ArrowRight)
            moveCursorRight();
        else if (keyType == KeyType.Enter)
            doEnter();
        else if (keyType == KeyType.Backspace)
            doBackspace();
        else if (keyType == KeyType.Delete)
            doDelete();
        else if (keyType == KeyType.Character)
            doCharacter(keyStroke);
    }

    private synchronized void doControlKeyCharacter(KeyStroke keyStroke) {
        KeyType keyType = keyStroke.getKeyType();

        if (keyType == KeyType.Character)
            if (keyStroke.getCharacter() == 'd') {
                doEnter();
                finish();
            }
    }

    private synchronized void doEnter() {
        terminal.putCharacter('\n');
        currentLine += "\n";

        try {
            inputWriter.write(currentLine.getBytes());
            inputWriter.flush();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        currentLine = "";
        origin = terminal.getCursorPosition();
    }

    private synchronized void moveCursorLeft() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition))
            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));
    }

    private synchronized boolean isPossibleToMoveLeft(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) > 0;
    }

    private synchronized int distanceFromOrigin(TerminalPosition cursorPosition) {
        int cursorColumn = cursorPosition.getColumn();
        int cursorRow = cursorPosition.getRow();
        int originColumn = origin.getColumn();
        int originRow = origin.getRow();
        int totalColumns = terminal.getTerminalSize().getColumns();

        return cursorColumn - originColumn + (totalColumns * (cursorRow - originRow));
    }

    private synchronized void moveCursorRight() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition))
            advanceCursor(cursorPosition);
    }

    private synchronized void advanceCursor(TerminalPosition cursorPosition) {
        if (isCursorAtEndOfRow(cursorPosition))
            terminal.setCursorPosition(0, cursorPosition.getRow() + 1);
        else
            terminal.setCursorPosition(cursorPosition.withRelativeColumn(1));
    }

    private synchronized boolean isCursorAtEndOfRow(TerminalPosition cursorPosition) {
        return cursorPosition.getColumn() == terminal.getTerminalSize().getColumns() - 1;
    }

    private synchronized boolean isPossibleToMoveRight(TerminalPosition cursorPosition) {
        return distanceFromOrigin(cursorPosition) < currentLine.length();
    }

    private synchronized void doBackspace() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveLeft(cursorPosition)) {
            String remaining = currentLine.substring(distanceFromOrigin(cursorPosition), currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition) - 1) + remaining;

            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');

            if (cursorPosition.getColumn() == 0)
                terminal.setCursorPosition(terminal.getTerminalSize().getColumns(), cursorPosition.getRow() - 1);
            else
                terminal.setCursorPosition(cursorPosition.withRelativeColumn(-1));
        }
    }

    private synchronized void doDelete() {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = currentLine.substring(distanceFromOrigin(cursorPosition) + 1, currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            terminal.putCharacter(' ');
            terminal.setCursorPosition(cursorPosition);
        }
    }

    private synchronized void doCharacter(KeyStroke keyStroke) {
        TerminalPosition cursorPosition = terminal.getCursorPosition();

        if (isPossibleToMoveRight(cursorPosition)) {
            String remaining = keyStroke.getCharacter()
                               + currentLine.substring(distanceFromOrigin(cursorPosition), currentLine.length());
            currentLine = currentLine.substring(0, distanceFromOrigin(cursorPosition)) + remaining;

            for (char c : remaining.toCharArray())
                terminal.putCharacter(c);

            advanceCursor(cursorPosition);
        } else {
            terminal.putCharacter(keyStroke.getCharacter());
            currentLine += keyStroke.getCharacter();

            advanceCursor(cursorPosition);
        }
    }

    public void finish() {
        isFinished = true;

        try {
            inputWriter.close();
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

}
